#!/usr/bin/env python3
# coding: utf-8

import importlib
import json
import os

import json5
from volkanic.cmdline import remember_cwd
from volkanic.utils import query_attr


class CommandNotFound(KeyError):
    pass


# deprecated
class CommandConf:
    def __init__(self, commands):
        self.commands = dict(commands)
        self.commands.setdefault("_global", {})

    @classmethod
    def _from_file(cls, name, loader, default_dir=None):
        path = cls._locate(name, default_dir)
        return cls(loader(open(path)))

    @classmethod
    def from_yaml(cls, name, default_dir=None):
        import yaml

        ext = os.path.splitext(name)[1].lower()
        if ext not in [".yml", ".yaml"]:
            name += ".yml"
        return cls._from_file(name, yaml.safe_load, default_dir)

    @classmethod
    def from_json(cls, name, default_dir=None):
        return cls._from_file(name, json.load, default_dir)

    @classmethod
    def from_json5(cls, name, default_dir=None):
        return cls._from_file(name, json5.load, default_dir)

    @staticmethod
    def _locate(path, default_dir):
        paths = [path]
        if default_dir is not None:
            paths.append(os.path.join(default_dir, path))
        for path in paths:
            if os.path.isfile(path):
                return path
        raise FileNotFoundError(path)

    @staticmethod
    def _execute(params):
        # only 'module' is a must-have
        prefix = params.get("prefix", "")
        module = prefix + params["module"]
        call = params.get("call", "run")
        args = params.get("args", [])
        kwargs = params.get("kwargs", {})
        if not isinstance(args, (list, tuple)):
            args = [args]
        m = importlib.import_module(module)
        query_attr(m, *call.split("."))(*args, **kwargs)

    def __call__(self, cmd):
        if cmd not in self.commands:
            raise CommandNotFound(str(cmd))
        params = dict(self.commands["_global"])
        params.update(self.commands[cmd])
        with remember_cwd():
            os.chdir(params.get("cd", "."))
            self._execute(params)

    @classmethod
    def run(cls, prog=None, args=None, default_dir=None, **kwargs):
        from argparse import ArgumentParser

        kwargs.setdefault("description", "volkanic command-conf runner")
        parser = ArgumentParser(prog=prog, **kwargs)
        parser.add_argument("name", help="a YAML file")
        parser.add_argument(
            "key",
            nargs="?",
            default="default",
            help="a top-level key in the YAML file",
        )
        ns = parser.parse_args(args=args)
        cconf = cls.from_yaml(ns.name, default_dir)
        cconf(ns.key)



#!/usr/bin/env python3
# coding: utf-8

_closing_of_known_tags = {
    "a": 2,
    "img": 0,
    "script": 2,
    "link": 0,
    "source": 0,
    "audio": 0,
    "video": 0,
}

_url_attrname_of_known_tags = {
    "a": "href",
    "link": "href",
    "img": "src",
    "script": "src",
    "source": "src",
    "audio": "src",
    "video": "src",
}

_default_props_of_known_tags = {
    "img": {
        "alt": "",
    },
    "script": {
        "type": "text/javascript",
    },
    "link": {
        "rel": "stylesheet",
        "type": "text/css",
    },
}

_templates = {
    0: "<{tagname} {props}>",
    1: "<{tagname} {props} />",
    2: "<{tagname} {props}>{content}</{tagname}>",
}


def fmttag(name, content, props=None, closing=None):
    props_parts = []
    for k, v in (props or {}).items():
        p = '{}="{}"'.format(k, v)
        props_parts.append(p)
    props_string = " ".join(props_parts)
    ctx = {"tagname": name, "props": props_string}
    if closing is None:
        closing = _closing_of_known_tags.get(name, 2)
    if closing == 2:
        ctx["content"] = content
    return _templates[closing].format(**ctx)


def wrap(content, *names):
    """
    :param content: string
    :param names: e.g. ('html', 'body')
    :return: string
    """
    parts = []
    names = list(names)
    for x in names:
        parts.append("<{}>".format(x))

    parts.append(content)
    names.reverse()
    for x in names:
        parts.append("</{}>".format(x))
    return "".join(parts)


def wrap_url(url, tagname):
    props = _default_props_of_known_tags.get(tagname, {})
    props[_url_attrname_of_known_tags.get(tagname, "href")] = url
    return fmttag(tagname, "", props=props)


def run_wrap_url(_, args):
    import sys

    tagname = args[0] if args else "a"
    for line in sys.stdin:
        print(wrap_url(line.strip(), tagname))


def wrap_default(content):
    return wrap(content, "html", "body")


def table(rows, props=None):
    trs = []
    for row in rows:
        tr = "".join("<td>" + c + "</td>" for c in row)
        tr = "<tr>" + tr + "</tr>"
        trs.append(tr)
    return fmttag("table", "\n".join(trs), props)



#!/usr/bin/env python3
# coding: utf-8

import datetime
import re
from datetime import timedelta
from typing import Union


def parse_time(time_string: str):
    """
    >>> parse_time('00:03:02,521')
    datetime.timedelta(seconds=182, microseconds=521000)
    """
    regex = re.compile("^([0-9]+):([0-9]+):([0-9]+),([0-9]+)$")
    match = regex.match(time_string)
    if match is None:
        raise ValueError("Unparseable timestamp: {}".format(time_string))
    hrs, mins, secs, msecs = map(int, match.groups())
    return timedelta(hours=hrs, minutes=mins, seconds=secs, milliseconds=msecs)


def datetime_from_timestamp_epoch1601(ts_1601: float, unit: int = 0):
    """
    Args:
        ts_1601: time since 1 January 1601
        unit: 0 = second, 6 = microsecond, 7 = 100 nanosecond
    Returns:
        a datetime.datetime instance
    See Also:
        https://en.wikipedia.org/wiki/Epoch_(computing)#Notable_epoch_dates_in_computing
    """
    # https://stackoverflow.com/a/26118615/2925169
    unit_per_sec = 10**unit
    ts = ts_1601 / unit_per_sec - 11644473600
    return datetime.datetime.fromtimestamp(ts)


datetime_from_timestamp_1601 = datetime_from_timestamp_epoch1601


def fmt_today(fmt="-") -> Union[str, datetime.date]:
    return relative_date(0, fmt)


def relative_date(delta: int = 0, fmt="-") -> Union[str, datetime.date]:
    date = datetime.date.today()
    if delta:
        date += datetime.timedelta(days=delta)
    if fmt is None:
        return date
    elif fmt == "":
        return date.strftime("%Y%m%d")
    elif fmt == "-":
        return date.strftime("%Y-%m-%d")
    return date.strftime(fmt)



#!/usr/bin/env python3
# coding: utf-8

import os
import time
from os.path import join

# noinspection PyPackageRequirements
import yaml


class ContextFile(object):
    def __init__(self, path):
        self.path = path
        self._data = {}
        self._defaults = {}
        self._mtime = 0
        self._load_from_file()

    def _load_from_file(self):
        mtime = os.path.getmtime(self.path)
        if mtime - (self._mtime or 0) < 0.001:
            return
        data = yaml.safe_load(open(self.path))
        if not isinstance(data, dict):
            raise TypeError("the contextmap file is not dict-like")
        data.update(self._defaults)
        self._data = data
        self._mtime = mtime

    def get(self, key, default=None):
        return self._data.get(key, default)

    def pop(self, key, default=None):
        return self._data.pop(key, default)

    def get_keys(self):
        self._load_from_file()
        return list(self._data)

    def setdefault(self, key, default=None):
        try:
            return self._data[key]
        except LookupError:
            return self._defaults.setdefault(key, default)


class RealContextFile(ContextFile):
    def __init__(self, path, ttl=60):
        super(ContextFile, self).__init__(path)
        self._xtime = 0
        self.ttl = ttl

    def get(self, key, default=None):
        current_time = time.time()
        # expired
        if current_time > self._xtime:
            self._load_from_file()
            self._xtime = current_time + self.ttl
        return self._data.get(key, default)

    def setdefault(self, key, default=None):
        return self._data.setdefault(key, default)


class ContextDirectory(object):
    def __init__(self, dir_path):
        if not os.path.isdir(dir_path):
            raise ValueError("requires a directory")
        self._cache = {}
        self._mtime = {}
        self.path = dir_path

    def _load_from_file(self, key):
        """
        Update _mtime and _cache if yaml.safe_load() is called.
        """
        path = join(self.path, key + ".yml")
        if not os.path.isfile(path):
            return
        mtime = os.path.getmtime(path)

        # avoid float equality test
        if mtime - self._mtime.get(key, 0) < 0.001:
            try:
                return self._cache[key]
            except LookupError:
                pass
        section = yaml.safe_load(open(path))
        self._mtime[key] = mtime
        self._cache[key] = section
        return section

    def setdefault(self, key, default=None):
        # only set default when key is not available in FS
        return self._cache.setdefault(key, self.get(key, default))

    def get(self, key, default=None):
        try:
            return self._cache[key]
        except LookupError:
            return self._load_from_file(key) or default

    def get_keys(self):
        _keys = []
        for filename in os.listdir(self.path):
            name, ext = os.path.splitext(filename)
            if ext == ".yml":
                _keys.append(name)
        return _keys


class RealContextDirectory(ContextDirectory):
    def __init__(self, dir_path, ttl=60):
        super(RealContextDirectory, self).__init__(dir_path)
        self._xtime = {}
        self.ttl = ttl

    def get(self, key, default=None):
        """
        Update _xtime if _load_from_file() is called and ttl available.
        Do NOT touch _cache and _mtime in this method;
        it is the job of _load_from_file.
        """
        current_time = time.time()
        expire_time = self._xtime.get(key, 0)

        # expired
        if current_time > expire_time:
            section = self._load_from_file(key)
            self._xtime[key] = current_time + self.ttl
            return section
        return super(RealContextDirectory, self).get(key, default)

    def setdefault(self, key, default=None):
        section = self._cache.setdefault(key, self.get(key, default))
        self._xtime.setdefault(key, float("inf"))
        return section


def context_load(path, ttl=None):
    if os.path.isdir(path):
        if ttl is None:
            return ContextDirectory(path)
        else:
            return RealContextDirectory(path, ttl)
    elif os.path.isfile(path):
        if ttl is None:
            return ContextFile(path)
        else:
            return RealContextFile(path, ttl)
    else:
        raise ValueError("path must be a file or directory")



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

from volkanic.introspect import format_function_path


class CheckZone(object):
    def __init__(self, checker):
        self._checker = checker
        self._entries = []

    def register(self, *funcs):
        func = None
        for func in funcs:
            entry = func, tuple(), dict()
            self._entries.append(entry)
        return func

    def register_with_params(self, *args, **kwargs):
        def _decorator(*funcs):
            func = None
            for func in funcs:
                entry = func, args, kwargs
                self._entries.append(entry)
            return func

        return _decorator

    def conduct_checks(self, clear=True):
        for func, args, kwargs in self._entries:
            retval = func(*args, **kwargs)
            if self._checker(retval):
                continue
            p = format_function_path(func)
            m = "introspect failed at {}".format(p)
            raise RuntimeError(m)
        if clear:
            self._entries = []


default_zone = CheckZone(bool)


def register(*funcs):
    return default_zone.register(*funcs)


def register_with_params(*args, **kwargs):
    return default_zone.register_with_params(*args, **kwargs)


def conduct_checks(clear=True):
    return default_zone.conduct_checks(clear=clear)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import hashlib
import re
import zlib

from joker.stream.utils import checksum


def guess_hash_algorithm(digest):
    algo_map = {
        ("bin", 16): "md5",
        ("bin", 20): "sha1",
        ("bin", 28): "sha224",
        ("bin", 32): "sha256",
        ("bin", 48): "sha384",
        ("bin", 64): "sha512",
        ("hex", 32): "md5",
        ("hex", 40): "sha1",
        ("hex", 56): "sha224",
        ("hex", 64): "sha256",
        ("hex", 96): "sha384",
        ("hex", 128): "sha512",
    }
    if isinstance(digest, bytes):
        try:
            digest = digest.decode("utf-8")
        except Exception:
            return algo_map.get(("bin", len(digest)))
    if not re.match(r"[0-9A-Fa-f]+", digest):
        return
    return algo_map.get(("hex", len(digest)))


class HashedPath(object):
    """
    /data/web/config.ini;;f6da8154d73f954cdfebd04d6c76cff8
    for integrity-check purpose only
    """

    delimiter = ";;"

    def __init__(self, digest, algo, path):
        self.digest = digest
        self.algo = algo
        self.path = path

    @staticmethod
    def compute_hash(path, algo="md5"):
        try:
            return checksum(path, algo=algo).hexdigest()
        except IOError:
            return hashlib.new(algo, b"\0").hexdigest()

    @classmethod
    def generate(cls, path, algo="md5"):
        digest = cls.compute_hash(path, algo)
        return cls(digest, algo, path)

    @classmethod
    def parse(cls, h_path):
        if cls.delimiter not in h_path:
            return cls("", "md5", h_path)
        path, digest = h_path.split(cls.delimiter, 2)
        algo = guess_hash_algorithm(digest)
        return cls(digest, algo, path)

    def __str__(self):
        return "{}{}{}".format(self.path, self.delimiter, self.digest)

    def verify(self):
        if not self.digest:
            return True
        hx1 = self.generate(self.path, self.algo or "md5")
        return hx1.digest == self.digest


class Hasher(object):
    """
    Unified interface for hashlib & zlib hash functions
    """

    __hfuncs__ = {
        "adler32": zlib.adler32,
        "crc32": zlib.crc32,
        "md5": hashlib.md5,
        "sha1": hashlib.sha1,
        "sha224": hashlib.sha224,
        "sha256": hashlib.sha256,
        "sha384": hashlib.sha384,
        "sha512": hashlib.sha512,
    }

    __hlens__ = {
        32: "md5",
        40: "sha1",
        56: "sha224",
        64: "sha256",
        96: "sha384",
        128: "sha512",
    }

    def __init__(self, name):
        self.name = name.lower()
        if self.name not in self.__hfuncs__:
            raise ValueError("hash func '{}' not found".format(name))
        if name in {"adl32", "crc32"}:
            self.hash = getattr(zlib, name)  # a checksum function
            self.emulate = True
            self.result = self.hash(bytes())  # intermediate result
        else:
            self.hash = getattr(hashlib, name)()  # a hash object
            self.emulate = False
            self.result = 0  # not used for hashlib functions

    def __repr__(self):
        cn = self.__class__.__name__
        return "{}({})".format(cn, repr(self.name))

    def _update_with_bytes(self, data):
        """
        :param data: (bytes)
        """
        # hashlib.md* / hashlib.sha*
        if not self.emulate:
            self.hash.update(data)
            return
        # zlib.adler32 / zlib.crc32
        if self.result is None:
            self.result = self.hash(data)
        else:
            self.result = self.hash(data, self.result)

    def _update_with_traverser(self, data):
        # from joker.nested.traverse import Traverser
        # traverser = Traverser(self, lambda x: self.update(x))
        # traverser()
        raise NotImplementedError

    def update(self, data):
        if isinstance(data, str):
            return self._update_with_bytes(data.encode())
        elif isinstance(data, bytes):
            return self._update_with_bytes(data)
        self._update_with_traverser(data)

    def hexdigest(self):
        # hashlib funcs
        if not self.emulate:
            return self.hash.hexdigest()
        # zlib.adler32 / zlib.crc32
        return hex(self.result & 0xFFFFFFFF).replace("0x", "")



#!/usr/bin/env python3
# coding: utf-8

import csv
import itertools
import sys

from volkanic.utils import load_json5_file


def dicts_to_rows(dicts: list, keys: list):
    return [[d.get(k) for k in keys] for d in dicts]


def rows_to_dicts(rows: list, keys: list):
    return [dict(zip(keys, r)) for r in rows]


def _write_to_csv(rows, path: str = None):
    if path is None:
        wr = csv.writer(sys.stdout)
        wr.writerows(rows)
        return
    with open(path, "a") as fout:
        wr = csv.writer(fout)
        wr.writerows(rows)


def dicts_to_csv(dicts: list, keys: list, path: str = None):
    if not dicts:
        return
    if not keys:
        keys = list(dicts[0].keys())
    rows = dicts_to_rows(dicts, keys)
    _write_to_csv(itertools.chain([keys], rows), path)


def json_to_csv(path: str, keys: list = None):
    data = load_json5_file(path)
    if not isinstance(data, list):
        raise TypeError(f"a list is required, got {type(data)}")
    if not data:
        return


def jsonl_to_csv(path: str, keys: list = None):
    data = load_json5_file(path)
    if not isinstance(data, list):
        raise TypeError(f"a list is required, got {type(data)}")
    if not data:
        return
    keys = list(data[0].keys())
    rows = dicts_to_rows(data, keys)
    wr = csv.writer(sys.stdout)
    wr.writerow(keys)
    wr.writerows(rows)


if __name__ == "__main__":
    json_to_csv(sys.argv[1])



#!/usr/bin/env python3
# coding: utf-8

import collections
import math


def compute_entropy(chars):
    n = len(chars)
    counter = collections.Counter(chars)
    probs = [ni / n for ni in counter.values()]
    return -sum([p * math.log2(p) for p in probs])



#!/usr/bin/env python3
# coding: utf-8


from __future__ import unicode_literals

import base64
import codecs
import functools
import itertools
import json
import os


# replaced by volkanic.utils.indented_json_print
def indented_json_print_legacy(obj, **kwargs):
    # https://stackoverflow.com/a/12888081/2925169
    decoder = codecs.getdecoder("unicode_escape")
    print_kwargs = {}
    for k in ["sep", "end", "file", "flush"]:
        if k in kwargs:
            print_kwargs[k] = kwargs.pop(k)
    kwargs.setdefault("indent", 4)
    s = json.dumps(obj, **kwargs)
    print(decoder(s)[0], **print_kwargs)


def search_with(regexes, text):
    for regex in regexes:
        mat = regex.search(text)
        if mat:
            return mat.groupdict()
    return dict()


def smart_split(text, sep):
    if isinstance(sep, str):
        parts = text.split(sep, 1)
        if len(parts) > 1:
            parts.insert(1, sep)
        return strip_all(*parts)
    # regex split
    return [p.strip() for p in sep.split(text, 1)]


def strip_all(*parts):
    return [p.strip() for p in parts]


def castable(func):
    """
    NOT supported: negative index (castfunc=-1 to get last item)
    >>> @castable
    ... def myfunc(*args):
    ...     for i in range(*args):
    ...         yield i
    ...
    >>> myfunc(12, castfunc=tuple)
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
    >>> myfunc(0, 12, 2, castfunc=2)
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)

    Purely syntax sugar,
    to make interactive use of some functions easier.
    Cast a generator function to list, set, or select n-th item, etc.

        myfunc(..., castfunc=list)   <=>  list(myfunc(...))
        myfunc(..., castfunc=1)      <=>  list(myfunc(...))[1]
    """

    @functools.wraps(func)
    def _decorated_func(*args, **kwargs):
        castfunc = None
        if "castfunc" in kwargs:
            castfunc = kwargs["castfunc"]
            del kwargs["castfunc"]

        retval = func(*args, **kwargs)
        if castfunc:
            if callable(castfunc):
                return castfunc(retval)
            # shortcut to pick up nth record
            if isinstance(castfunc, int):
                return next(itertools.islice(retval, castfunc, None))
            return TypeError("castfunc must be a callable or integer")
        return retval

    return _decorated_func


# https://docs.python.org/3/library/enum.html#using-automatic-values
class AttrEchoer(object):
    """
    Resembles an enum type
    Reduces typos by using syntax based completion of dev tools

    Example:

        @instanciate_with_foolproof
        class Event(AttrEchoer):
            _prefix = 'event'
            bad_params = ''  # assign whatever
            unauthorized_access = ''
            undefined_fault = ''
            ...

        assert Event.unauthoried  == 'event.bad_params'
    """

    _prefix = "_root."

    def __init__(self):
        pass

    def __getattribute__(self, key):
        kls = type(self)
        if key in kls.__dict__ and key != "_prefix":
            if not kls._prefix:
                return key
            return "{}{}".format(kls._prefix, key)
        return object.__getattribute__(self, key)


def locate_standard_conf(package):
    # a string is also acceptable
    name = getattr(package, "__name__", package)
    workdirs = [os.environ.get("{}_WORKDIR".format(name.upper()))]

    # look for ~/.<package> and /data/<package>
    # ONLY IF <package>_WORKDIR is not set
    if workdirs[0] is None:
        workdirs = [
            os.path.expanduser("~/.{}".format(name)),
            "/data/{}/".format(name),
        ]
        workdirs = [d for d in workdirs if os.path.isdir(d)]

    for d in workdirs:
        p = os.path.join(d, "configs/{}.yml".format(name))
        if p and os.path.isfile(p):
            return p
    raise ValueError("config file not found")


def gen_random_string(length):
    n = 1 + int(length * 0.625)
    return base64.b32encode(os.urandom(n)).decode()[:length]


def setup_basic_logging(level="INFO"):
    import logging

    params = {
        "level": level,
        "format": "%(asctime)s %(levelname)s %(name)s %(message)s",
        "datefmt": "%Y%m%d~%H:%M:%S",
    }
    logging.basicConfig(**params)


def strip_lines(lines, lstrip=True, rtrip=True):
    stripfuncs = []
    if rtrip:
        stripfuncs.append(str.lstrip)
    if lstrip:
        stripfuncs.append(str.rsplit)
    if not stripfuncs:
        stripfuncs = [lambda s: s.rstrip("\n")]
    elif len(stripfuncs) == 2:
        stripfuncs = [str.strip]
    for line in lines:
        stripfuncs[0](line)



#!/usr/bin/env python3
# coding: utf-8

import os
from collections import deque


def iter_path_bisections(path: str):
    names = deque([])
    while path:
        path, name = os.path.split(path)
        if not path or not name:
            break
        names.appendleft(name)
        yield path, os.path.join(*names)


class PathMapper:
    def __init__(self, fs_path: str, user_path: str):
        self._real_prefix = os.path.realpath(fs_path)
        self._user_prefix = user_path

    def userpath(self, *paths):
        return os.path.join(self._user_prefix, *paths)

    def realpath(self, *paths):
        return os.path.join(self._real_prefix, *paths)

    def under(self, *paths, mkdirs=False):
        path = self.realpath(*paths)
        if mkdirs:
            dir_ = os.path.split(path)[0]
            os.makedirs(dir_, exist_ok=True)
        return path

    def exists(self, *paths):
        return os.path.exists(self.realpath(*paths))

    def isfile(self, *paths):
        return os.path.isfile(self.realpath(*paths))

    def isdir(self, *paths):
        return os.path.isdir(self.realpath(*paths))



#!/usr/bin/env python3
# coding: utf-8

import base64
import binascii
import urllib.parse
from urllib.parse import ParseResult


def parse_qs_flat(query: str):
    d = urllib.parse.parse_qs(query, keep_blank_values=True)
    return {k: v[0] for k, v in d.items()}


def parse_url(url) -> (ParseResult, dict):
    pr = urllib.parse.urlparse(url)
    d = parse_qs_flat(pr.query)
    return pr, d


def parse_url_to_dict(url, component_prefix="@") -> dict:
    pr = urllib.parse.urlparse(url)
    d = parse_qs_flat(pr.query)
    # noinspection PyProtectedMember
    d.update({component_prefix + k: v for k, v in pr._asdict().items()})
    return d


def _base32(bs):
    return base64.b32encode(bs).decode().replace("=", "").lower()


def _base64_encode(s):
    return base64.urlsafe_b64encode(s.encode()).decode().replace("=", "")


def _base64_decode(s):
    padded = s[::-1] + "=" * (-len(s) % 4)
    return base64.urlsafe_b64decode(padded).decode()


def url_obfuscate(s):
    ss = base64.urlsafe_b64encode(s.encode())
    return ss.decode().replace("=", "")[::-1]


def url_deobfuscate(s):
    try:
        padded = s[::-1] + "=" * (-len(s) % 4)
        return base64.urlsafe_b64decode(padded).decode()
    except (UnicodeDecodeError, binascii.Error):
        pass



#!/usr/bin/env python3
# coding: utf-8

import dataclasses

from volkanic.compat import cached_property


def prefetch_cached_properties(obj):
    d = getattr(obj, "__dict__", {})
    for key, val in d.items():
        if isinstance(val, cached_property):
            getattr(obj, key)


class NovelDict(dict):
    def __getattr__(self, key: str):
        return self[key]

    def __setattr__(self, key: str, value):
        self[key] = value

    def __delattr__(self, key: str):
        del self[key]

    def get_properties(self) -> dict:
        props = {}
        for key, val in self.__class__.__dict__.items():
            if isinstance(val, (property, cached_property)):
                props[key] = getattr(self, key)
        return props

    def get_data_and_properties(self) -> dict:
        data = self.copy()
        data.update(self.get_properties())
        return data


@dataclasses.dataclass
class Record:
    @classmethod
    def from_dict(cls, dic: dict):
        names = set([f.name for f in dataclasses.fields(cls)])
        kwargs = {k: v for k, v in dic.items() if k in names}
        # noinspection PyArgumentList
        return cls(**kwargs)



#!/usr/bin/env python3
# coding: utf-8

import dataclasses
from functools import cached_property

from pymongo.collection import Collection


@dataclasses.dataclass
class Comparison:
    c1_values: list
    c2_values: list

    @classmethod
    def from_collections(cls, c1: Collection, c2: Collection, field: str):
        args = (
            [r[field] for r in c1.find(projection=[field])],
            [r[field] for r in c2.find(projection=[field])],
        )
        return cls(*args)

    @cached_property
    def values(self):
        return self.c1_values + self.c2_values

    @cached_property
    def uniq_values(self):
        return self.uniq_c1_values | self.uniq_c2_values

    @cached_property
    def uniq_c1_values(self):
        return set(self.c1_values)

    @cached_property
    def uniq_c2_values(self):
        return set(self.c2_values)

    @cached_property
    def common_values(self):
        vals = [v for v in self.c1_values if v in self.uniq_c2_values]
        return [v for v in self.c2_values if v in self.uniq_c1_values] + vals

    @cached_property
    def uniq_common_values(self):
        return self.uniq_c1_values & self.uniq_c2_values

    def find_diff(self):
        return (
            self.uniq_c1_values - self.uniq_c2_values,
            self.uniq_c2_values - self.uniq_c1_values,
        )

    @property
    def counts(self):
        return (
            len(self.values),
            len(self.c1_values),
            len(self.c2_values),
            len(self.common_values),
            len([v for v in self.c1_values if v not in self.uniq_c2_values]),
            len([v for v in self.c2_values if v not in self.uniq_c1_values]),
        )

    @property
    def uniq_counts(self):
        return (
            len(self.uniq_values),
            len(self.uniq_c1_values),
            len(self.uniq_c2_values),
            len(self.uniq_common_values),
            len(self.uniq_c1_values - self.uniq_c2_values),
            len(self.uniq_c2_values - self.uniq_c1_values),
        )

    @property
    def duplication_ratio(self):
        return [t / u for t, u in zip(self.counts, self.uniq_counts)]

    def get_stats(self):
        attrnames = ["counts", "uniq_counts", "duplication_ratio"]
        return {k: getattr(self, k) for k in attrnames}



#!/usr/bin/env python3
# coding: utf-8

import re


class RegexTester(object):
    def __init__(self, regexes=None, strings=None):
        self.regexes = regexes or []
        self.strings = strings or []

    def check_all_regexes(self, string):
        return [r.search(string) for r in self.regexes]

    def check_all_strings(self, pattern):
        return [re.search(pattern, s) for s in self.strings]


RegexGroup = RegexTester



#!/usr/bin/env python3
# coding: utf-8

import os

import volkanic.environ
from volkanic.utils import (
    under_home_dir, abs_path_join, abs_path_join_and_mkdirs
)


class GlobalInterface(volkanic.environ.GlobalInterfaceTrial):
    package_name = 'm14.meta'

    def under_data_dir(self, *paths, mkdirs=False):
        if 'data_dir' not in self.conf:
            names = self.package_name.split('.')
            data_dir = os.path.join('/data/local', *names)
            self.conf.setdefault('data_dir', data_dir)
        return super().under_data_dir(*paths, mkdirs=mkdirs)

    @classmethod
    def under_m14_dir(cls, *paths, mkdirs=False):
        base_dir = os.environ.get('JOKER_HOME') or under_home_dir('.m14')
        if not mkdirs:
            return abs_path_join(base_dir, *paths)
        return abs_path_join_and_mkdirs(base_dir, *paths)

    @classmethod
    def under_m14_subdir(cls, *paths, mkdirs=False):
        _paths = cls.namespaces[1:]
        _paths.extend(paths)
        return cls.under_m14_dir(*_paths, mkdirs=mkdirs)

    @classmethod
    def _get_conf_path_names(cls):
        names = cls.namespaces.copy()
        names.append(cls._get_option('confpath_filename'))
        return names



#!/usr/bin/env python3
# coding: utf-8

import os.path
import sys

# noinspection PyPackageRequirements
import yaml
from volkanic.utils import under_home_dir

_conf = None


def _load_conf():
    paths = [under_home_dir(".m14-default.yml"), "/etc/m14-default.yml"]
    for path in paths:
        if os.path.isfile(path):
            return yaml.safe_load(open(path))


def _get_default_prefix():
    if sys.platform.startswith("win"):
        return r"c:\data"
    return "/data"


def _get_conf():
    global _conf
    if _conf is None:
        _conf = _load_conf() or {}
    if "default" not in _conf:
        _conf["default"] = _get_default_prefix()
    return _conf


def under_default_dir(package, *paths):
    conf = _get_conf()
    name = getattr(package, "__name__", str(package)).split(".")[-1]
    try:
        dir_ = conf[name]
    except LookupError:
        dir_ = os.path.join(conf.get("default"), name)
    return os.path.join(dir_, *paths)



#!/usr/bin/env python3
# coding: utf-8
import os
import re

import setuptools
from setuptools import find_namespace_packages

# import joker; exit(1)
# DO NOT import your package from your setup.py

namespace = "m14"
package_name = "meta"
description = "experimental utilities for m14 projects"


def read(filename):
    with open(filename) as f:
        return f.read()


def version_find():
    if namespace:
        path = "{}/{}/__init__.py".format(namespace, package_name)
    else:
        path = "{}/__init__.py".format(package_name)
    root = os.path.dirname(__file__)
    path = os.path.join(root, path)
    regex = re.compile(r"""^__version__\s*=\s*('|"|'{3}|"{3})([.\w]+)\1\s*(#|$)""")
    # print('path', path)
    with open(path) as fin:
        for line in fin:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            mat = regex.match(line)
            if mat:
                return mat.groups()[1]
    raise ValueError("__version__ definition not found")


config = {
    "name": "m14",
    "version": version_find(),
    "description": "" + description,
    "keywords": "m14",
    "url": "",
    "author": "frozflame",
    "author_email": "frozflame@outlook.com",
    "license": "GNU General Public License (GPL)",
    "packages": find_namespace_packages(include=["m14.*"]),
    "namespace_packages": ["m14"],
    "zip_safe": False,
    "install_requires": read("requirements.txt"),
    "classifiers": [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    # ensure copy static file to runtime directory
    "include_package_data": True,
    "long_description": read("README.md"),
    "long_description_content_type": "text/markdown",
}

setuptools.setup(**config)



#!/usr/bin/env python3
# coding: utf-8

import json
import logging
import os.path
import re
import shlex
import sys
from functools import lru_cache, wraps

import volkanic.utils

# TODO: drop this
from joker.textmanip.path import make_new_path
from volkanic.introspect import razor


def dump_json_request_to_curl(method: str, url: str, data=None, aslist=False):
    method = method.upper()
    if method == "GET":
        parts = ["curl", url]
    else:
        parts = [
            "curl",
            "-X",
            method,
            url,
            "-H",
            "Content-Type: application/json",
            "-d",
            json.dumps(razor(data), ensure_ascii=False),
        ]
    if aslist:
        return parts
    parts = [shlex.quote(s) for s in parts]
    return " ".join(parts)


@lru_cache(1024)
def _load_json5_file_cached(path: str):
    return volkanic.utils.load_json5_file(path)


def load_json5_file_cached(path: str):
    path = os.path.abspath(path)
    return _load_json5_file_cached(path)


get_json_config = load_json5_file_cached


def copy_fields(record: dict, keys: list, keymap: dict, default=None):
    new_record = {k: record.get(k, default) for k in keys}
    for old_key, new_key in keymap.items():
        new_record[new_key] = record.get(old_key, default)
    return new_record


def camel_case_split(name: str):
    # https://stackoverflow.com/a/29920015/2925169
    matches = re.finditer(
        r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)", name
    )
    return [m.group(0) for m in matches]


def camel_case_to_underscore(name: str):
    return "_".join(s.lower() for s in camel_case_split(name))


def text_routine_catstyle(func, *paths):
    # routine function for quick and dirty text manipulation scripts
    if not paths:
        paths = sys.argv[1:]

    if not paths:
        for line in sys.stdin:
            print(func(line))
        return
    for path in paths:
        for line in open(path):
            print(func(line))


def text_routine_perfile(func, ext=".out"):
    # routine function for quick and dirty text manipulation scripts
    for path in sys.argv[1:]:
        outpath = make_new_path(path, ext)
        func(path, outpath)


def chained_bracket(record: dict, *keys):
    current_rec = record
    for k in keys:
        try:
            current_rec = current_rec[k]
        except (KeyError, IndexError):
            current_rec = {}
    return current_rec


def dotget(record: dict, key: str):
    return chained_bracket(record, key.split("."))


def _get_instance_from_method_args(func, args: tuple):
    if not args:
        return
    instance = args[0]
    try:
        raw_func = instance.__class__.__dict__.get(func.__name__).raw_func
    except (AttributeError, TypeError):
        return
    if raw_func != func:
        return
    return instance


def once_per_instance(func):
    """Runs a method (successfully) only once per instance."""

    @wraps(func)
    def _resulting_method(*args, **kwargs):
        self = _get_instance_from_method_args(func, args)
        d = getattr(self, "__dict__", {})
        if d.get(func.__qualname__):
            msg = f"method {func.__qualname__} can only be called once"
            raise RuntimeError(msg)
        rv = func(*args, **kwargs)
        d[func.__qualname__] = 1
        return rv

    _resulting_method.raw_func = func
    return _resulting_method


def assert_equal(a, b):
    assert a == b, (a, b)


# deprecated. use joker.environ.utils.check_{inclusive,exclusive}_* functions
def check_prefixes(string: str, include=None, exclude=None) -> bool:
    """
    Args:
        string (str): string to be checked
        include: None = include anything
        exclude: None = no exlusion rule
    """
    if exclude is not None:
        for prefix in exclude:
            if string.startswith(prefix):
                return False
    if include is not None:
        for prefix in include:
            if string.startswith(prefix):
                return True
        return False
    else:
        return True


def easylog(level):
    root = logging.root
    fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    formatter = logging.Formatter(fmt)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    root.setLevel(level)
    root.addHandler(handler)


def read_lines(path: str):
    for line in open(path):
        line = line.strip()
        if not line:
            continue
        yield line


# deprecated: use joker.textmanip.random_hex
# Python 3.5+
if hasattr(bytes, "hex"):

    def random_hex(length=12):
        return os.urandom(length).hex()

else:

    def random_hex(size=12):
        import base64

        b = os.urandom(size)
        return base64.b16encode(b).decode("ascii")
